﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2018 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Collections.Concurrent;
using System.Collections.Generic;
using System.Linq;
using System.Threading;
using System.Threading.Tasks;
using HEAL.Attic;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Optimization;
using HeuristicLab.Parameters;

namespace HeuristicLab.Problems.Modifiers {
  [StorableType("7DE16061-1E13-4039-A78F-E33821F41648")]
  [Item("EvaluationRepetitionProblemModifier", "A problem modifier that repeats the evaluation multiple times and averages the results. If you use the parallelization feature, make sure the underlying problem is thread-safe")]
  public class EvaluationRepetitionProblemModifier : ProblemModifier {
    public const string RepetitionIdVariableName = "RepetitionID";

    #region Parameter Names
    public const string RepetitionsParameterName = "Repetitions";
    private const string ParallelismParameterName = "Degree of Parallelism";
    private const string EnforceParameterName = "Enforce Parallelism";
    #endregion

    #region Parameters
    public IFixedValueParameter<IntValue> RepetitionsParameter => (IFixedValueParameter<IntValue>)Parameters[RepetitionsParameterName];
    public IFixedValueParameter<IntValue> ParallelismParameter => (IFixedValueParameter<IntValue>)Parameters[ParallelismParameterName];
    public IFixedValueParameter<BoolValue> EnforceParameter => (IFixedValueParameter<BoolValue>)Parameters[EnforceParameterName];
    #endregion

    #region Parameter Properties
    public int Repetitions {
      get => RepetitionsParameter.Value.Value;
      set => RepetitionsParameter.Value.Value = value;
    }
    public int Parallelism {
      get => ParallelismParameter.Value.Value;
      set => ParallelismParameter.Value.Value = value;
    }
    public bool Enforce {
      get => EnforceParameter.Value.Value;
      set => EnforceParameter.Value.Value = value;
    }
    #endregion

    #region Constructors & Cloning
    [StorableConstructor]
    protected EvaluationRepetitionProblemModifier(StorableConstructorFlag _) : base(_) { }
    protected EvaluationRepetitionProblemModifier(EvaluationRepetitionProblemModifier original, Cloner cloner) : base(original, cloner) { }
    public EvaluationRepetitionProblemModifier() {
      Parameters.Add(new FixedValueParameter<IntValue>(RepetitionsParameterName, "The number of repetitions per evaluation.", new IntValue(10)));
      Parameters.Add(new FixedValueParameter<IntValue>(ParallelismParameterName, "The degree of parallelism", new IntValue(1)));
      Parameters.Add(new FixedValueParameter<BoolValue>(EnforceParameterName, "Whether the degree of parallelism is enforced (thread parallel) or just a maximum (C# parallel foreach)", new BoolValue(false)));
    }
    [StorableHook(HookType.AfterDeserialization)]
    private void AfterDeserialization() {
    }
    public override IDeepCloneable Clone(Cloner cloner) {
      return new EvaluationRepetitionProblemModifier(this, cloner);
    }
    #endregion

    #region ProblemModifier
    public override double[] ModifiedEvaluate(Individual individual, IRandom random) {
      var iBag = new ConcurrentBag<IResult>();
      var rBag = new ConcurrentBag<Tuple<string, double[]>>();
      var options = new ParallelOptions() { MaxDegreeOfParallelism = Parallelism };

      if (!Enforce)
        Parallel.ForEach(Enumerable.Range(0, Repetitions), options, repetition => {
          EvaluateRep(repetition, iBag, rBag, individual, random);
        });
      else {
        var threads = new List<Thread>();
        foreach (var repetition in Enumerable.Range(0, Repetitions)) {
          var t = new Thread(() => { EvaluateRep(repetition, iBag, rBag, individual, random); });
          threads.Add(t);
          t.Start();
        }
        foreach (var t in threads) t.Join();
      }

      individual["SimulationQualities"] = new ResultCollection(iBag.OrderBy(x => x.Name));
      individual["SimulationResult"] = new ResultCollection(rBag.Select(t => new Result(t.Item1, new DoubleArray(t.Item2))).OrderBy(x => x.Name));
      return Enumerable.Range(0, Maximization.Length).Select(i => rBag.Select(x => x.Item2[i]).Average()).ToArray();
    }
    #endregion

    #region Helpers
    private void EvaluateRep(int repetition, ConcurrentBag<IResult> iBag, ConcurrentBag<Tuple<string, double[]>> rBag, Individual individual, IRandom random) {
      var extendedIndividual = individual.Copy();
      extendedIndividual[RepetitionIdVariableName] = new IntValue(repetition);
      var qualities = base.ModifiedEvaluate(extendedIndividual, random);
      iBag.Add(new Result("Repetition " + repetition, ExtractNewVariables(individual, extendedIndividual)));
      rBag.Add(Tuple.Create("Repetition " + repetition, qualities));
    }

    private static ResultCollection ExtractNewVariables(Individual oldInd, Individual newInd) {
      var dict = oldInd.Values.ToDictionary(x => x.Key, x => x.Value);
      return new ResultCollection(newInd.Values.Where(v1 => !dict.ContainsKey(v1.Key)).Select(v1 => new Result(v1.Key, v1.Value)));
    }
    #endregion
  }
}
